/* ****** Perspective System *****
**
***  Indira Vidyaprakash
**  July 16 1996
********************************** */


/* A set of classes to parse, represent and display 3D wireframe models
   represented in Wavefront .obj format. */

import java.applet.Applet;
import java.awt.*;
import java.io.*;
import java.net.*;


/** The representation of a 3D model */
class PerspModel3D extends Canvas {

  // Will be user defined values. But for the first case these should match values in
  // Main Program
  
  double tox = 0.0;
  double toy = 0.0;
  double toz = 0.0;
  
  double fromx = 50.0;
  double fromy = 0.0;
  double fromz = 0.0;
  
  double upx = 0.0;
  double upy = 0.0; // this is the approx. up vector
  double upz = 1.0;

  // Need to get the  values of d ,f and h
  double d = 30.0;
  double f = 70.0;
  double h = 20.0;
    
  
  // For the View Vert and the Persp vert
  double view_vert[];
  double persp_vert[];
  
  double xfac;
  int prevx, prevy;
  double xtheta, ytheta;
  double scalefudge = 1.0;
  
  boolean painted = true;
  
  double vert[];
  int tvert[];
  int nvert, maxvert;
  int con[];
  int ncon, maxcon;
  boolean transformed;
  Matrix3D mat;

  double f_vert[];
  int ft_vert[];
  int f_line[];
  
  Matrix3D tmat = new Matrix3D();
  Matrix3D amat = new Matrix3D();
  
  double xmin, xmax, ymin, ymax, zmin, zmax;

  Dimension minSize;
  int width, height;
  
  // For Clipping
 // For clipping
  boolean above1 = false;
  boolean below1 = false;
  boolean left1 = false;
  boolean right1 = false;
  boolean behind1 = false;
  boolean front1 = false;
  boolean above2 = false;
  boolean below2 = false;
  boolean left2 = false;
  boolean right2 = false;
  boolean behind2 = false;
  boolean front2 = false;

public static final int ABOVE = 1;
public static final int BELOW = 2;
public static final int RIGHT = 3;
public static final int LEFT = 4;
public static final int FRONT = 5;
public static final int BEHIND = 6;

  double clip_vert[];
  int clip_con[];
  
  int ccon;     // Number of lines in the clip_con array
  int cpvert;  // Number of vertices in the clip_vert array
  
  //-----------


  
  PerspModel3D () {
    mat = new Matrix3D ();
    
    // Need to set the height and the width of the canvas
    width = 325;
    height = 200;
    resize(width, height);
    
    minSize = new Dimension(width, height);
    
  }
  
  
public Dimension preferredSize() {
  return minimumSize();
}
  
public synchronized Dimension minimumSize() {
  return minSize;
}
  
  
  /** Create a 3D model by parsing an input stream */
  PerspModel3D (URL isurl, String objName) throws IOException, MalformedURLException {
        
    this();
    
    try{
      
      InputStream is = null;
      is = new URL(isurl, objName).openStream();
      
      StreamTokenizer st = new StreamTokenizer(is);
      st.eolIsSignificant(true);
      st.commentChar('#');
    scan:
      while (true) {
	switch (st.nextToken()) {
	  default:
	    break scan;
	  case StreamTokenizer.TT_EOL:
	    break;
	  case StreamTokenizer.TT_WORD:
	    if ("v".equals(st.sval)) {
	      double x = 0.0, y = 0.0, z = 0.0;
	      if (st.nextToken() == StreamTokenizer.TT_NUMBER) {
		x = st.nval;
		if (st.nextToken() == StreamTokenizer.TT_NUMBER) {
		  y = st.nval;
		  if (st.nextToken() == StreamTokenizer.TT_NUMBER)
		    z = st.nval;
		}
	      }
	      addVert((double) x, (double) y, (double) z);
	      while (st.ttype != StreamTokenizer.TT_EOL &&
		     st.ttype != StreamTokenizer.TT_EOF)
		st.nextToken();
	    } else if ("f".equals(st.sval) || "fo".equals(st.sval) || "l".equals(st.sval)) {
	      int start = -1;
	      int prev = -1;
	      int n = -1;
	      while (true)
		if (st.nextToken() == StreamTokenizer.TT_NUMBER) {
		  n = (int) st.nval;
		  if (prev >= 0)
		    add(prev - 1, n - 1);
		  if (start < 0)
		    start = n;
		  prev = n;
		} else if (st.ttype == '/')
		  st.nextToken();
		else
		  break;
	      if (start >= 0)
		add(start - 1, prev - 1);
	      if (st.ttype != StreamTokenizer.TT_EOL)
		break scan;
	    } else {
	      while (st.nextToken() != StreamTokenizer.TT_EOL
		     && st.ttype != StreamTokenizer.TT_EOF);
	    }
	}
	}
      is.close();
      if (st.ttype != StreamTokenizer.TT_EOF)
	throw new FileFormatException(st.toString());
	
      
      compress();
       
    } catch (Exception e) { }
    
    
  }
  
  /** Add a vertex to this model */
  int addVert(double x, double y, double z) {
    int i = nvert;
    if (i >= maxvert)
      if (vert == null) {
	maxvert = 100;
	vert = new double[maxvert * 3];
      } else {
	maxvert *= 2;
	double nv[] = new double[maxvert * 3];
	System.arraycopy(vert, 0, nv, 0, vert.length); // src, dest
	vert = nv;
      }
    i *= 3;
    vert[i] = x;
    vert[i + 1] = y;
    vert[i + 2] = z;
    return nvert++;
  }

  /** Add a line from vertex p1 to vertex p2 */
  void add(int p1, int p2) {
    int i = ncon;
    if (p1 >= nvert || p2 >= nvert)
      return;
    if (i >= maxcon)
      if (con == null) {
	maxcon = 100;
	con = new int[maxcon];
      } else {
	maxcon *= 2;
	int nv[] = new int[maxcon];
	System.arraycopy(con, 0, nv, 0, con.length);
	con = nv;
      }
    if (p1 > p2) {
      int t = p1;
      p1 = p2;
      p2 = t;
    }
    con[i] = (p1 << 16) | p2;
    ncon = i + 1;
  }
  
  /** Transform all the points in this model */
  void transform() {
    if (transformed || cpvert <= 0)
      return;
    if (tvert == null || tvert.length < cpvert * 3)
      tvert = new int[500];
    mat.transform(persp_vert, tvert, cpvert);
    transformed = true;
  }

  
  /* Quick Sort implementation
   */
private void quickSort(int a[], int left, int right)
  {
    int leftIndex = left;
    int rightIndex = right;
    int partionElement;
    if ( right > left)
    {
      
      /* Arbitrarily establishing partition element as the midpoint of
       * the array.
       */
      partionElement = a[ ( left + right ) / 2 ];
      
      // loop through the array until indices cross
      while( leftIndex <= rightIndex )
      {
	/* find the first element that is greater than or equal to 
	 * the partionElement starting from the leftIndex.
	 */
	while( ( leftIndex < right ) && ( a[leftIndex] < partionElement ) )
	  ++leftIndex;
	
	/* find an element that is smaller than or equal to 
	 * the partionElement starting from the rightIndex.
	 */
	while( ( rightIndex > left ) && 
	       ( a[rightIndex] > partionElement ) )
	  --rightIndex;
	
	// if the indexes have not crossed, swap
	if( leftIndex <= rightIndex ) 
	{
	  swap(a, leftIndex, rightIndex);
	  ++leftIndex;
	  --rightIndex;
	}
      }
      
      /* If the right index has not reached the left side of array
       * must now sort the left partition.
       */
      if( left < rightIndex )
	quickSort( a, left, rightIndex );
      
      /* If the left index has not reached the right side of array
       * must now sort the right partition.
       */
      if( leftIndex < right )
	quickSort( a, leftIndex, right );
      
    }
  }

  
private void swap(int a[], int i, int j)
  {
    int T;
    T = a[i]; 
    a[i] = a[j];
    a[j] = T;
  }
  
  
  /** eliminate duplicate lines */
  void compress() {
    int limit = ncon;
    int c[] = con;
    quickSort(con, 0, ncon - 1);
    int d = 0;
    int pp1 = -1;
    for (int i = 0; i < limit; i++) {
      int p1 = c[i];
      if (pp1 != p1) {
	c[d] = p1;
	d++;
      }
      pp1 = p1;
    }
    ncon = d;
  }
  
   

    /** Paint this model to a graphics context.  It uses the matrix associated
	with this model to map from model space to screen space.
	The next version of the browser should have double buffering,
	which will make this *much* nicer */
  
    public void paint(Graphics g) {

      tView();

      Clipping();
      // Calculate the number of vertices in the clip_vert array
      // Value is obtained form ccon( number of lines in the line array
      cpvert = ccon * 2;
      
      tPersp();

      // Tp draw the frustrum surrounding the Persp Projection
      pFrustrum();
      
      mat.unit();
            
      // Might have to translate to middle of the Drawing Area (Canvas) rather than Boundbox
      mat.translate(-(xmin + xmax) / 2,-(ymin + ymax) / 2,-(zmin + zmax) / 2);

      mat.mult(amat);
      
      mat.scale(xfac, -xfac, 16 * xfac / width);
           
      // Translating to the center
      mat.translate(width / 2, height / 2, 8);
      transformed = false;


      
      // Frustrum Drawing

      // First transform the points
      mat.transform(f_vert, ft_vert, 8);

      // Have attempted to paint the near plane in an yellow color and
      // the far plane in a diff color
      for (int i = 0; i < 8; i++) {
      int T = f_line[i];
      int p1 = ((T >> 16) & 0xFFFF) * 3;
      int p2 = (T & 0xFFFF) * 3;
      
	g.setColor(new Color(255,255,0));
	g.drawLine(ft_vert[p1], ft_vert[p1 + 1], ft_vert[p2], ft_vert[p2 + 1]);
    }     

      // far Plane
      for (int i = 8; i < 12; i++) {
      int T = f_line[i];
      int p1 = ((T >> 16) & 0xFFFF) * 3;
      int p2 = (T & 0xFFFF) * 3;
      
	g.setColor(new Color(0,180,210));
	g.drawLine(ft_vert[p1], ft_vert[p1 + 1], ft_vert[p2], ft_vert[p2 + 1]);
    }     

      // Label the canvas
    g.setColor(new Color(255,255,255));          
    g.drawString("Screen Coordinates", 250,20);


    // Draw the Object 
    if (persp_vert == null || cpvert <= 0){
       setPainted();
	return;
    }
      transform();
      
      int lg = 0;
      int lim = ccon;
      int c[] = clip_con;
      int v[] = tvert;
      if (lim <= 0 || nvert <= 0){
	 setPainted();
	return;
      }
      
      
      for (int i = 0; i < lim; i++) {
	int T = c[i];
	int p1 = ((T >> 16) & 0xFFFF) * 3;
	int p2 = (T & 0xFFFF) * 3;
	
	g.setColor(new Color(255,0,0));
	g.drawLine(v[p1], v[p1 + 1],
		   v[p2], v[p2 + 1]);
      }


      setPainted();

      
    }
  
    
  // Method to do Tview
    
  void tView(){

    // should get inputs for look to, look from and camera up vector from USER
    
     // Using the above values to calculate the U, V and N and also Camera Position C
     
     double ux, uy, uz;
     double vx, vy, vz;
     double nx, ny, nz;
     double cx, cy, cz;

     // Temp variables to normalize the up vector
     double nvx, nvy, nvz;
     

     // Computing Camera Position as look from position
     cx = fromx;
     cy = fromy;
     cz = fromz;
     
     // Computing N as NORM(to - from)
     // First computing NORM
     double norm = (double) Math.sqrt(((tox-fromx) * (tox-fromx)) + ((toy-fromy)*(toy-fromy)) + ((toz-fromz)*(toz-fromz)));
     
     nx = (tox-fromx) / norm;
     ny = (toy-fromy) / norm;
     nz = (toz-fromz) / norm;
     
     // Computing NORM (UP)
     double norm_up = (double) Math.sqrt((upx*upx) + (upy*upy) + (upz*upz));
     
     // computing norm(UP) dot N
     double dot = ((upx /norm_up)*nx) + ((upy/norm_up)*ny) + ((upz/norm_up)*nz);
     
     // Computing (norm(UP) - ( norm(UP) dot N) N)
     nvx = (upx /norm_up) - (dot * nx);
     nvy = (upy /norm_up) - (dot * ny);
     nvz = (upz /norm_up) - (dot * nz);
     
     // Obtaining the norm of the up vector
     double norm_v = (double) Math.sqrt((nvx*nvx) + (nvy*nvy) + (nvz*nvz));
     
     // Computing V as = NORM (norm(UP) - ( norm(UP) dot N) N)
     vx = nvx / norm_v;
     vy = nvy / norm_v;
     vz = nvz / norm_v;
     
         
     // Computing U as N cross V
     ux = (ny*vz) - (nz*vy);
     uy = (nz*vx) - (nx*vz); 
     uz = (nx*vy) - (ny*vx);

     // Since we use a 4x3 matrix it is of the form
     //           ux               vx                nx
     //           uy               vy                ny
     //           uz               vz                nz
     //           -cxux-cyuy-czuz  -cxvx-cyvy-czvz   -cxnx-cyny-cznz
     //

     // Calculating the last line of the matrix so that it is easier later
     double new_cx = -(cx*ux)-(cy*uy)-(cz*uz);
     double new_cy = -(cx*vx)-(cy*vy)-(cz*vz);
     double new_cz = -(cx*nx)-(cy*ny)-(cz*nz);

     // Initilaize all the time
     // Changed
     if ( view_vert == null)
       view_vert = new double[300];
     
     // Multiply vert (for nvert number of vertices) into the View Matrix

     for (int i = nvert * 3; (i -= 3) >= 0;) {
       double x = vert[i];
       double y = vert[i + 1];
       double z = vert[i + 2];

       // Changed
       view_vert[i]     = -(x * ux + y * uy + z * uz + new_cx);
       view_vert[i + 1] = (x * vx + y * vy + z * vz + new_cy);
       view_vert[i + 2] = (x * nx + y * ny + z * nz + new_cz);
     }
  }
  // Method to do Persepective Projection

  void tPersp(){

    

     // Initilaize all the time
    
     if ( persp_vert == null)
       persp_vert = new double[500];
    
    // Operate on the vert Matrix but only after cranking it through the View Matrix and Clipping
        
      for (int i = cpvert * 3; (i -= 3) >= 0;) {
	// changed
       double x = clip_vert[i];
       double y = clip_vert[i + 1];
       double z = clip_vert[i + 2];
       
       // Changed
       persp_vert[i]     = ((d * x) / (h * z));
       persp_vert[i + 1] = ((d * y) / (h * z));
       persp_vert[i + 2] = ((f / (f-d)) - ((f*d)/((f-d)*z)));
      }


    double v[] = persp_vert;
    double xmin = v[0], xmax = xmin;
    double ymin = v[1], ymax = ymin;
    double zmin = v[2], zmax = zmin;
    for (int i = cpvert * 3; (i -= 3) > 0;) {
      double x = v[i];
      if (x < xmin)
	xmin = x;
      if (x > xmax)
	xmax = x;
      double y = v[i + 1];
      if (y < ymin)
	ymin = y;
      if (y > ymax)
	ymax = y;
      double z = v[i + 2];
      if (z < zmin)
	zmin = z;
      if (z > zmax)
	zmax = z;
    }
    this.xmax = xmax;
    this.xmin = xmin;
    this.ymax = ymax;
    this.ymin = ymin;
    this.zmax = zmax;
    this.zmin = zmin;


   
    
  }

  void pFrustrum(){

    
    f_vert = new double[24];
     
    // Fill up the Array
    
     
 
    // Vertex 1 - lower right corner
    f_vert[0] = 1.0;
    f_vert[1] = -1.0;
    f_vert[2] = 0.0;
    
    // Upper right
    f_vert[3] = 1.0;
    f_vert[4] = 1.0;
    f_vert[5] = 0.0;
    
    // Upper Left
    f_vert[6] = -1.0;
    f_vert[7] = 1.0;
    f_vert[8] = 0.0;
    
    // Vertex 4 - lower left corner
    f_vert[9] = -1.0;
    f_vert[10] = -1.0;
    f_vert[11] = 0.0;
    
        
    // Vertex 5 - back right corner
    f_vert[12] = 1.0;
    f_vert[13] = -1.0;
    f_vert[14] = 1.0;
    
    f_vert[15] = 1.0;
    f_vert[16] = 1.0;
    f_vert[17] = 1.0;
    
    f_vert[18] = -1.0;
    f_vert[19] = 1.0;
    f_vert[20] = 1.0;
    
    // Vertex 8
    f_vert[21] = -1.0;
    f_vert[22] = -1.0;
    f_vert[23] = 1.0;
    
    
    // Array of Lines to be joined
    f_line = new int[12];
    
    
    f_line[0] = (0 << 16) | 1;
    f_line[1] = (1 << 16) | 2;
    f_line[2] = (2 << 16) | 3;
    f_line[3] = (0 << 16) | 3;
    f_line[4] = (0 << 16) | 4;
    f_line[5] = (1 << 16) | 5;
    f_line[6] = (2 << 16) | 6;
    f_line[7] = (3 << 16) | 7;
    f_line[8] = (4 << 16) | 5;
    f_line[9] = (5 << 16) | 6;
    f_line[10] = (6 << 16) | 7;
    f_line[11] = (4 << 16) | 7;
    

    // Find the bounding box of the frustrum and apply translation accordingly so that it
    // matches with the view
    double vf[] = f_vert;
    for (int i = 24; (i -= 3) > 0;) {
      double x = vf[i];
      if (x < this.xmin)
	this.xmin = x;
      if (x > this.xmax)
	this.xmax = x;
      double y = vf[i + 1];
      if (y < this.ymin)
	this.ymin = y;
      if (y > this.ymax)
	this.ymax = y;
      double z = vf[i + 2];
      if (z < this.zmin)
	this.zmin = z;
      if (z > this.zmax)
	this.zmax = z;
    }

    // Computing the factor for scaling - this is done only here since we always do Frustrum after
    // Persp object
    double xw = xmax - xmin;
    double yw = ymax - ymin;
    double zw = zmax - zmin;
    if (yw > xw)
      xw = yw;
    if (zw > xw)
      xw = zw;
    double f1 = width / xw;
    double f2 = height / xw;
    this.xfac = 0.7f * (f1 < f2 ? f1 : f2) * scalefudge;

    // New Transformation Matrix - to hold the integer values after transformation
    ft_vert = new int[24];
    
  }


  // For Clipping the Object. Object is clipped in View Coordinates and only then transformed
  // to perspective projection
void Clipping() {

    // The inputs for this will be the view_vert since clipping is done in View Coordinate
    // And also the array of line joining one point to another


    int lim = ncon;
    // Making a copy of the original line array
    clip_con = new int[300];
    //System.arraycopy(con, 0, clip_con, 0, con.length); 
    

    // For the clipped lines connect
    ccon = 0;
    
    boolean accept = false;
    boolean reject = false;

    boolean inside1 = false;
    boolean inside2 = false;

   
    ClipPt clip = new ClipPt();

    // always a new clip_vert
    clip_vert = new double[300];
        
    // Looping through all the lines for the length of the line array
    for(int i = 0; i < lim; i++)
    {
      // Get the two lines
      int T = con[i];
      int p1 = ((T >> 16) & 0xFFFF) * 3;  // First Vertex
      int p2 = (T & 0xFFFF) * 3;          // Second Vertex

      // This will be in order of the clip_vert array
      int j = ccon;
      
    
      // Finding the index of the clip_vert array 
      int v1 = j*6;
      int v2 = j*6+3;
       // Assign values to clip_vert here from view_ve
      clip_vert[v1] = view_vert[p1];
      clip_vert[v1+1] = view_vert[p1+1];
      clip_vert[v1+2] = view_vert[p1+2];

      clip_vert[v2] = view_vert[p2];
      clip_vert[v2+1] = view_vert[p2+1];
      clip_vert[v2+2] = view_vert[p2+2];

      //Assigning clip_con the new values
      clip_con[j] = ( (v1/3) << 16 ) | (v2/3);

          
      // Incrementing the ccon value
      ccon = ccon + 1;
      

      while(!(accept || reject)){

	
	CheckVertex(clip_vert[v1], clip_vert[v1+1], clip_vert[v1+2], 1);
	inside1 = !(above1 || below1 || right1 || left1 || front1 || behind1);
	
	
	CheckVertex(clip_vert[v2], clip_vert[v2+1], clip_vert[v2+2], 2);
	inside2 = !(above2 || below2 || right2 || left2 || front2 || behind2);
	
	// To accept line inside view volume
	accept = inside1 && inside2;

	if(!accept){
	  
	  // To check if reject outside the view volume
	  reject = ((above1 && above2) || (below1 && below2) || (right1 && right2) ||
		    (left1 && left2) || (behind1 && behind2) || (front1 && front2));

	  if (reject){
	  

	   // Only need to decrement the value of ccon
	   ccon = ccon - 1;
	  }

	  // This is if not trivial accept or reject -> CLIP
	  if (!reject){
	   
	    if(!inside1){

	      if(above1){
		
		IntersectPoint(v1,v2,ABOVE,clip);
		CopyPoint(v1,clip);
	      }
	      else if(below1){
		
		IntersectPoint(v1, v2,BELOW,clip);
		CopyPoint(v1,clip);
	      }
	      else if(left1){
		
		IntersectPoint(v1,v2,LEFT,clip);
		CopyPoint(v1,clip);
	      }
		else if(right1){
		
		IntersectPoint(v1, v2,RIGHT,clip);
		CopyPoint(v1,clip);
	      }
	      else if(front1){
		
		IntersectPoint(v1, v2,FRONT,clip);
		CopyPoint(v1,clip);
	      }
	      else {
		
		IntersectPoint(v1,v2,BEHIND,clip);
		CopyPoint(v1,clip);
		
	      }
	    }
	    else{
	      if(above2){
		
		IntersectPoint(v2, v1,ABOVE,clip);
		CopyPoint(v2,clip);	       
	      }
	      else if(below2){
		
		IntersectPoint(v2, v1,BELOW,clip);
		CopyPoint(v2,clip);
	      }
	      else if(left2){
		
		IntersectPoint(v2,v1,LEFT,clip);
		CopyPoint(v2,clip);
	      }
		else if(right2){
		
		IntersectPoint(v2,v1,RIGHT,clip);
		CopyPoint(v2,clip);
	      }
	      else if(front2){
		
		IntersectPoint(v2,v1,FRONT,clip);
		CopyPoint(v2,clip);
	      }
	      else {
		
		IntersectPoint(v2,v1,BEHIND,clip);
		CopyPoint(v2,clip);
	      }

	    }
	 
	  }// !reject 

	 
       }// !accept
	 
      }// a/r while
      accept = false;
      reject = false;
   
    }// line loop

   
  }
  

  void CheckVertex(double x, double y, double z, int vertNum){
 

    //vertNum indicates whether it will be first or second point
    if (vertNum == 1) {


      above1 =  y > ((h*z)/d + 0.1);
      below1 = y <  (-(h*z)/d - 0.1);
      left1 = x < ((-(h*z)/d) - 0.1);
      right1 = x > ((h*z)/d + 0.1);
      front1 = z < d;
      behind1 = z > f;
            
    }else {

      above2 = y > ((h*z)/d + 0.1);
      below2 = y <  (-(h*z)/d - 0.1);
      left2 = x < (-(h*z)/d - 0.1);
      right2 = x > ((h*z)/d + 0.1);
      front2 = z < d;
      behind2 = z > f;
            
    }

  }

  void IntersectPoint(int pt1, int pt2, int type, ClipPt clip){

   
    // Find the Intersection Point - use clip_vert
    double t = 0.0;

    
    double x1 = clip_vert[pt1];
    double y1 = clip_vert[pt1 + 1];
    double z1 = clip_vert[pt1 + 2];

    double x2 = clip_vert[pt2];
    double y2 = clip_vert[pt2 + 1];
    double z2 = clip_vert[pt2 + 2];

      
    switch (type){
      case ABOVE:
	if (( t = (y2 - y1) - ((h/d)*(z2-z1)) ) != 0.0) {
	  t = (z1 - y1) /t;
	  // Calculating the intersection point
	  clip.ix = (x2-x1)*t + x1;
	  clip.iy = (y2-y1)*t + y1;
	  clip.iz = (d/h)*clip.iy;

	
	}
	break;
	
      case BELOW:
	if (( t = (y2 - y1) + ((h/d)*(z2-z1)) ) != 0.0) {
	  t = ((-(h/d)*z1) - y1) /t;
	  // Calculating the intersection point
	  clip.ix = (x2-x1)*t + x1;
	  clip.iy = (y2-y1)*t + y1;
	  clip.iz = -(d/h)*clip.iy;

	  
	}
	break;
	
      case RIGHT:
	if (( t = (x2 - x1) - ((h/d)*(z2-z1)) ) != 0.0) {
	  t = ((h/d)*z1 - x1) /t;
	  // Calculating the intersection point
	  clip.ix = (x2-x1)*t + x1;
	  clip.iy = (y2-y1)*t + y1;
	  clip.iz = (d/h)*clip.ix;

	 
	}
	break;
	
      case LEFT:
	if (( t = (x2 - x1) + ((h/d)*(z2-z1)) ) != 0.0) {
	  t = ((-(h/d)*z1) - x1) /t;
	  // Calculating the intersection point
	  clip.ix = (x2-x1)*t + x1;
	  clip.iy = (y2-y1)*t + y1;
	  clip.iz = -(d/h)*clip.ix;

	  
	}
	break;

      case FRONT:
	if( z2 != z1){
	  t = (d - z1) / (z2 - z1);
	  clip.ix = (x2-x1)*t + x1;
	  clip.iy = (y2-y1)*t + y1;
	  clip.iz = d;
	  
	 
	}
	break;

      case BEHIND:
	if( z2 != z1){
	  t = (f - z1) / (z2 - z1);
	  clip.ix = (x2-x1)*t + x1;
	  clip.iy = (y2-y1)*t + y1;
	  clip.iz = f;

	 
	}
	break;
	
	  }
   
  }

  

  void CopyPoint(int pt, ClipPt clip){

    clip_vert[pt] = clip.ix;
    clip_vert[pt + 1] = clip.iy;
    clip_vert[pt + 2] = clip.iz;
   
  }
  

  // For the mouse action, should be able to drag both the frustrum
  //and the view object together

public boolean mouseDown(Event e, int x, int y) {
  prevx = x;
  prevy = y;
  return true;
}
  
public boolean mouseDrag(Event e, int x, int y) {
  tmat.unit();
  double xtheta = (prevy - y) * 360.0f / width;
  double ytheta = (x - prevx) * 360.0f / height;
  tmat.xrot(xtheta);
  tmat.yrot(ytheta);
  amat.mult(tmat);
  if (painted) {
    painted = false;
    repaint();
  }
  prevx = x;
  prevy = y;
  return true;
}



private synchronized void setPainted() {
  painted = true;
  notifyAll();
}

  
}


	 

  



