/* ******************* VIEW COORDINATE SYSTEM ************************************
 **
 **   Indira Vidyaprakash
 **   July 16 1996
 **   To draw the Viewing Frustrum and the object in the View co-ordinate system
 **
 ******************************************************************************** */



/* A set of classes to parse, represent and display 3D wireframe models
   in the View Coordinate System
   represented in Wavefront .obj format. */

import java.awt.*;
import java.applet.Applet;
import java.io.*;
import java.net.*;



/** The representation of a 3D model */
class ViewModel3D extends Canvas {

  double fromx = 50.0;
  double fromy = 0.0;
  double fromz = 0.0;

  // These values come from the MAINPROGRAM
  
  // should get inputs for look to, look from and camera up vector from USER
  double tox = 0.0;
  double toy = 0.0;
  double toz = 0.0;
  
  double upx = 0.0;
  double upy = 0.0; // this is the approx. up vector
  double upz = 1.0;

  // This is to draw the frustrum
  double d = 30.0;
  double f = 70.0;
  double h = 20.0;

   
  boolean painted = true;
  double xfac;
  double xtheta, ytheta;
  int prevx, prevy;
  
  double scalefudge = 1.0;

  // For the Rotations

  Matrix3D tmat = new Matrix3D();
  Matrix3D amat = new Matrix3D();
  
  // Added for the Frustrum
  double f_vert[];
  int ft_vert[];
  int f_line[];
  
   // This is for the Array that will hold all the View transformed Points that will finally be
   // drawn
   // Changed
   double view_vert[];
   
   double vert[];
  int tvert[];
  int nvert, maxvert;
  int con[];
  int ncon, maxcon;
  boolean transformed;
  Matrix3D mat;
 
  
  double xmin, xmax, ymin, ymax, zmin, zmax;

  Dimension minSize;
  int width, height;

  ViewModel3D () {
    mat = new Matrix3D ();
  
    
    // Need to set the height and the width of the canvas
    width = 325;
    height = 200;
    resize(width, height);
      
      minSize = new Dimension(width, height);
    
  }

  
public Dimension preferredSize () {
  return minimumSize();
}
  
public synchronized Dimension minimumSize() {
  return minSize;
}

  
  /** Create a 3D model by parsing an input stream */
  /** Might have to pass the Mainprogram also form Draw Canvas */
  ViewModel3D (URL isurl, String objName) throws IOException, MalformedURLException {
    
    this();

    try{

      InputStream is = null;
      is = new URL(isurl, objName).openStream();
      
      StreamTokenizer st = new StreamTokenizer(is);
      st.eolIsSignificant(true);
      st.commentChar('#');
    scan:
      while (true) {
	switch (st.nextToken()) {
	  default:
	    break scan;
	  case StreamTokenizer.TT_EOL:
	    break;
	  case StreamTokenizer.TT_WORD:
	    if ("v".equals(st.sval)) {
	      double x = 0.0, y = 0.0, z = 0.0;
	      if (st.nextToken() == StreamTokenizer.TT_NUMBER) {
		x = st.nval;
		if (st.nextToken() == StreamTokenizer.TT_NUMBER) {
		  y = st.nval;
		  if (st.nextToken() == StreamTokenizer.TT_NUMBER)
		    z = st.nval;
		}
	      }
	      addVert((double) x, (double) y, (double) z);
	      while (st.ttype != StreamTokenizer.TT_EOL &&
		     st.ttype != StreamTokenizer.TT_EOF)
		st.nextToken();
	    } else if ("f".equals(st.sval) || "fo".equals(st.sval) || "l".equals(st.sval)) {
	      int start = -1;
	      int prev = -1;
	      int n = -1;
	      while (true)
		if (st.nextToken() == StreamTokenizer.TT_NUMBER) {
		  n = (int) st.nval;
		  if (prev >= 0)
		    add(prev - 1, n - 1);
		  if (start < 0)
		    start = n;
		  prev = n;
		} else if (st.ttype == '/')
		  st.nextToken();
		else
		  break;
	      if (start >= 0)
		add(start - 1, prev - 1);
	      if (st.ttype != StreamTokenizer.TT_EOL)
		break scan;
	    } else {
	      while (st.nextToken() != StreamTokenizer.TT_EOL
		     && st.ttype != StreamTokenizer.TT_EOF);
	    }
	}
      }
      is.close();
      if (st.ttype != StreamTokenizer.TT_EOF)
	throw new FileFormatException(st.toString());
      
      

      compress();
    } catch (Exception e) { }
    
  }
  
  
  /** Add a vertex to this model */
  int addVert(double x, double y, double z) {
    int i = nvert;
    if (i >= maxvert)
      if (vert == null) {
	maxvert = 100;
	vert = new double[maxvert * 3];
      } else {
	maxvert *= 2;
	double nv[] = new double[maxvert * 3];
	System.arraycopy(vert, 0, nv, 0, vert.length); // src, dest
	vert = nv;
      }
    i *= 3;
    vert[i] = x;
    vert[i + 1] = y;
    vert[i + 2] = z;
    return nvert++;
  }

  
  /** Add a line from vertex p1 to vertex p2 */
  void add(int p1, int p2) {
    int i = ncon;
    if (p1 >= nvert || p2 >= nvert)
      return;
    if (i >= maxcon)
      if (con == null) {
	maxcon = 100;
	con = new int[maxcon];
      } else {
	maxcon *= 2;
	int nv[] = new int[maxcon];
	System.arraycopy(con, 0, nv, 0, con.length);
	con = nv;
      }
    if (p1 > p2) {
      int t = p1;
      p1 = p2;
      p2 = t;
    }
    con[i] = (p1 << 16) | p2;
    ncon = i + 1;
  }

  
  /** Transform all the points in this model */
  void transform() {
   
    if (transformed || nvert <= 0)
      return;
    if (tvert == null || tvert.length < nvert * 3)
      tvert = new int[nvert*3];
    mat.transform(view_vert, tvert, nvert);  // Changed
    transformed = true;
  }

  
  /* Quick Sort implementation
   */
private void quickSort(int a[], int left, int right)
  {
    int leftIndex = left;
    int rightIndex = right;
    int partionElement;
    if ( right > left)
    {
      
      /* Arbitrarily establishing partition element as the midpoint of
       * the array.
       */
      partionElement = a[ ( left + right ) / 2 ];
      
      // loop through the array until indices cross
      while( leftIndex <= rightIndex )
      {
	/* find the first element that is greater than or equal to 
	 * the partionElement starting from the leftIndex.
	 */
	while( ( leftIndex < right ) && ( a[leftIndex] < partionElement ) )
	  ++leftIndex;
	
	/* find an element that is smaller than or equal to 
	 * the partionElement starting from the rightIndex.
	 */
	while( ( rightIndex > left ) && 
	       ( a[rightIndex] > partionElement ) )
	  --rightIndex;
	
	// if the indexes have not crossed, swap
	if( leftIndex <= rightIndex ) 
	{
	  swap(a, leftIndex, rightIndex);
	  ++leftIndex;
	  --rightIndex;
	}
      }
      
      /* If the right index has not reached the left side of array
       * must now sort the left partition.
       */
      if( left < rightIndex )
	quickSort( a, left, rightIndex );
      
      /* If the left index has not reached the right side of array
       * must now sort the right partition.
       */
      if( leftIndex < right )
	quickSort( a, leftIndex, right );
      
    }
  }

  
private void swap(int a[], int i, int j)
  {
    int T;
    T = a[i]; 
    a[i] = a[j];
    a[j] = T;
  }
  
  
  /** eliminate duplicate lines */
  void compress() {
    int limit = ncon;
    int c[] = con;
    quickSort(con, 0, ncon - 1);
    int d = 0;
    int pp1 = -1;
    for (int i = 0; i < limit; i++) {
      int p1 = c[i];
      if (pp1 != p1) {
	c[d] = p1;
	d++;
      }
      pp1 = p1;
    }
    ncon = d;
  }
  
  

    /** Paint this model to a graphics context.  It uses the matrix associated
	with this model to map from model space to screen space. * */
  // Does the paint for both the viewing frustrum as well as the object 
	
    public void paint(Graphics g) {

      tView();
      tFrustrum();
  
      mat.unit();
           
      // Check to see how this works out
      mat.translate(-(xmin + xmax) / 2,-(ymin + ymax) / 2,-(zmin + zmax) / 2);
      
      mat.mult(amat);
      
      mat.scale(xfac, -xfac, 16 * xfac / width);

           
      // Translating to the center
      mat.translate(width / 2, height / 2, 8);
      
      transformed = false;
      	     
      
      if (view_vert == null || nvert <= 0)   // Changed
	return;
      
      
      transform();
      
      int lg = 0;
      int lim = ncon;
      int c[] = con;
      int v[] = tvert;
      if (lim <= 0 || nvert <= 0)
	return;
      
      
      for (int i = 0; i < lim; i++) {
	int T = c[i];
	int p1 = ((T >> 16) & 0xFFFF) * 3;
	int p2 = (T & 0xFFFF) * 3;
	
	g.setColor(new Color(255,0,0));
	g.drawLine(v[p1], v[p1 + 1],v[p2], v[p2 + 1]);
      }

   // Frustrum Drawing

      mat.transform(f_vert, ft_vert, 12);

      for (int i = 0; i < 12; i++) {
      int T = f_line[i];
      int p1 = ((T >> 16) & 0xFFFF) * 3;
      int p2 = (T & 0xFFFF) * 3;
      
	g.setColor(new Color(255,255,0));
	g.drawLine(ft_vert[p1], ft_vert[p1 + 1], ft_vert[p2], ft_vert[p2 + 1]);
    }

      // Coloring the back plane in a different color
      // to match with the Persp Projection Screen
      for (int i = 12; i < 16; i++) {
      int T = f_line[i];
      int p1 = ((T >> 16) & 0xFFFF) * 3;
      int p2 = (T & 0xFFFF) * 3;
      
	g.setColor(new Color(0,180,210));
	g.drawLine(ft_vert[p1], ft_vert[p1 + 1], ft_vert[p2], ft_vert[p2 + 1]);
    }
      
      //Painting the Upvector Dir in a diff color along with the arrows
      for (int i = 16; i < 19; i++) {
      int T = f_line[i]; 
      int p1 = ((T >> 16) & 0xFFFF) * 3;
      int p2 = (T & 0xFFFF) * 3;
      
	g.setColor(new Color(0,255,0));
	g.drawLine(ft_vert[p1], ft_vert[p1 + 1], ft_vert[p2], ft_vert[p2 + 1]);

      }
      // Label the canvas
      g.setColor(new Color(255,255,255));          
      g.drawString("View Coordinates", 250,20);
	
      setPainted();
}

  // Method to do Tview
    
  void tView(){

  
    // Using the above values to calculate the U, V and N and also Camera Position C

     double ux, uy, uz;
     double vx, vy, vz;
     double nx, ny, nz;
     double cx, cy, cz;

     // Temp variables to normalize the up vector
     double nvx, nvy, nvz;
     
     // Computing Camera Position as look from position
     cx = fromx;
     cy = fromy;
     cz = fromz;
     
     // Computing N as NORM(to - from)
     // First computing NORM
     double norm_n = (double) Math.sqrt(((tox-fromx) * (tox-fromx)) + ((toy-fromy)*(toy-fromy)) + ((toz-fromz)*(toz-fromz)));

     nx = (tox-fromx) / norm_n;
     ny = (toy-fromy) / norm_n;
     nz = (toz-fromz) / norm_n;

     // Computing NORM (UP)
     double norm_up = (double) Math.sqrt((upx*upx) + (upy*upy) + (upz*upz));
     
     // computing norm(UP) dot N
     double dot = ((upx /norm_up)*nx) + ((upy/norm_up)*ny) + ((upz/norm_up)*nz);

     // Computing (norm(UP) - ( norm(UP) dot N) N)
     nvx = (upx /norm_up) - (dot * nx);
     nvy = (upy /norm_up) - (dot * ny);
     nvz = (upz /norm_up) - (dot * nz);
     
     // Obtaining the norm of the up vector
     double norm_v = (double) Math.sqrt((nvx*nvx) + (nvy*nvy) + (nvz*nvz));
     
     // Computing V as = NORM (norm(UP) - ( norm(UP) dot N) N)
     vx = nvx / norm_v;
     vy = nvy / norm_v;
     vz = nvz / norm_v;
     
          
     // Computing U as N cross V
     ux = (ny*vz) - (nz*vy);
     uy = (nz*vx) - (nx*vz); 
     uz = (nx*vy) - (ny*vx);

     // Since we use a 4x3 matrix it is of the form
     //           ux               vx                nx
     //           uy               vy                ny
     //           uz               vz                nz
     //           -cxux-cyuy-czuz  -cxvx-cyvy-czvz   -cxnx-cyny-cznz
     //

     // Calculating the last line of the matrix so that it is easier later
     double new_cx = -(cx*ux)-(cy*uy)-(cz*uz);
     double new_cy = -(cx*vx)-(cy*vy)-(cz*vz);
     double new_cz = -(cx*nx)-(cy*ny)-(cz*nz);


     // Initialize view_vert always here
     // Changed
      if (view_vert == null) {
	view_vert = new double[300];
      } 
     
     // Multiply vert (for nvert number of vertices) into the View Matrix and store in view_vert

     for (int i = nvert * 3; (i -= 3) >= 0;) {
       double x = vert[i];
       double y = vert[i + 1];
       double z = vert[i + 2];

       // Changed
       view_vert[i]     = -(x * ux + y * uy + z * uz + new_cx);
       view_vert[i + 1] = (x * vx + y * vy + z * vz + new_cy);
       view_vert[i + 2] = (x * nx + y * ny + z * nz + new_cz);
     }


     // Find the bounding box for the model only now so that you will get the View vertices
     
     double v[] = view_vert;
     double xmin = v[0], xmax = xmin;
     double ymin = v[1], ymax = ymin;
     double zmin = v[2], zmax = zmin;
     
     for (int i = nvert * 3; (i -= 3) > 0;) {
       double x = v[i];
       if (x < xmin)
	 xmin = x;
       if (x > xmax)
	 xmax = x;
       double y = v[i + 1];
       if (y < ymin)
	 ymin = y;
       if (y > ymax)
	 ymax = y;
      
       //double z = v[i + 2];
       // if (z < zmin)
       //zmin = z;
       //if (z > zmax)
       //zmax = z;
	}
     this.xmax = xmax;
     this.xmin = xmin;
     this.ymax = ymax;
     this.ymin = ymin;
     this.zmax = zmax;
     this.zmin = zmin;

     
  }

  

  void tFrustrum() {
    
    f_vert = new double[36];
    double h1 = (h*f) / d;        // The height of the Far Plane
    
    // Fill up the Array
    
    // Look from point should be O
    f_vert[0] = 0.0;
    f_vert[1] = 0.0;
    f_vert[2] = 0.0;
    
    // For the corners of the near plane
    // Vertex 1 - lower right corner
    f_vert[3] = h;
    f_vert[4] = -h;
    f_vert[5] = d;
    
    // Upper right
    f_vert[6] = h ;
    f_vert[7] = h;
    f_vert[8] = d;
    
    // Upper Left
    f_vert[9] = -h;
    f_vert[10] =h;
    f_vert[11] = d;
    
    // Vertex 4 - lower left corner
    f_vert[12] = -h;
    f_vert[13] = -h;
    f_vert[14] = d;
    
    // For the corners of the far plane
    
    // Vertex 5
    f_vert[15] = h1 ;
    f_vert[16] = -h1;
    f_vert[17] = f;
    
    f_vert[18] = h1;
    f_vert[19] = h1;
    f_vert[20] = f;
    
    f_vert[21] = -h1;
    f_vert[22] = h1;
    f_vert[23] = f;
    
    // Vertex 8
    f_vert[24] = -h1;
    f_vert[25] = -h1;
    f_vert[26] = f;

    //Vertex 9 - Line for UP
    f_vert[27] = 0.0;
    f_vert[28] = 20.0;
    f_vert[29] = 0.0;

    //Vertex 10 - Arrows
    f_vert[30] = 4.0;
    f_vert[31] = 16.0;
    f_vert[32] = 0.0;
    
    f_vert[33] = -4.0;
    f_vert[34] = 16.0;
    f_vert[35] = 0.0;
    
    // Array of Lines to be joined
    f_line = new int[19];
    
    
    f_line[0] = (0 << 16) | 1;
    f_line[1] = (0 << 16) | 2;
    f_line[2] = (0 << 16) | 3;
    f_line[3] = (0 << 16) | 4;
    f_line[4] = (1 << 16) | 2;
    f_line[5] = (2 << 16) | 3;
    f_line[6] = (3 << 16) | 4;
    f_line[7] = (1 << 16) | 4;
    f_line[8] = (1 << 16) | 5;
    f_line[9] = (2 << 16) | 6;
    f_line[10] = (3 << 16) | 7;
    f_line[11] = (4 << 16) | 8;
    f_line[12] = (5 << 16) | 6;
    f_line[13] = (6 << 16) | 7;
    f_line[14] = (7 << 16) | 8;
    f_line[15] = (5 << 16) | 8;
    f_line[16] = (0 << 16) | 9;
    f_line[17] = (9 << 16) | 10;
    f_line[18] = (9 << 16) | 11;
    
    // Find the bounding box of the frustrum and apply translation accordingly so that it
    // matches with the view
    double vf[] = f_vert;
    for (int i = 36; (i -= 3) > 0;) {
      double x = vf[i];
      if (x < this.xmin)
	this.xmin = x;
      if (x > this.xmax)
	this.xmax = x;
      double y = vf[i + 1];
      if (y < this.ymin)
	this.ymin = y;
      if (y > this.ymax)
	this.ymax = y;
      double z = vf[i + 2];
      if (z < this.zmin)
	this.zmin = z;
      if (z > this.zmax)
	this.zmax = z;
    }

    // Computing the factor for scaling - this is done only here since we always do Frustrum after
    // View object
    double xw = xmax - xmin;
    double yw = ymax - ymin;
    double zw = zmax - zmin;
    if (yw > xw)
      xw = yw;
    if (zw > xw)
      xw = zw;
    double f1 = width / xw;
    double f2 = height / xw;
    this.xfac = 0.7f * (f1 < f2 ? f1 : f2) * scalefudge;

    // New Transformation Matrix - to hold the integer values after transformation
    ft_vert = new int[36];
    
  }


  // For the mouse action, should be able to drag both the frustrum
  //and the view object together

public boolean mouseDown(Event e, int x, int y) {
  prevx = x;
  prevy = y;
  return true;
}
  
public boolean mouseDrag(Event e, int x, int y) {
  tmat.unit();
  double xtheta = (prevy - y) * 360.0f / width;
  double ytheta = (x - prevx) * 360.0f / height;
  tmat.xrot(xtheta);
  tmat.yrot(ytheta);
  amat.mult(tmat);
  if (painted) {
    painted = false;
    repaint();
  }
  prevx = x;
  prevy = y;
  return true;
}



private synchronized void setPainted() {
  painted = true;
  notifyAll();
}


  
}

 
  
   
